// Simple tool for converting text to binary

#ifdef WIN32
#include <windows.h>
#include <io.h>
#endif

#include <velib/velib_config.h>
#include <velib/io/ve_vstream.h>
#include <velib/types/variant.h>
#include <velib/types/variant_console.h>
#include <velib/utils/ve_getopt.h>

#include <stdio.h>
#include <iostream>
#include <fcntl.h>
#include <string>
#include <string.h>

using namespace std;

static un8 buffer[255];
static VeStream stream;
static bool bigEndian = false;
static bool isHex = false;

void usage(char* name)
{
	printf("%s [type] [type] ..\r\n", name);
	printf("\r\n");
	printf("Convert binary data to textual representation.\r\n");
	printf("\r\n");
	printf("For supported types see -t.\r\n");
	printf("\r\n");
	printf("The binary data is read from stdin.\r\n");
	printf("The output is formatted like the input of bin2var.\r\n");
	printf("\r\n");
	printf("OPTIONS: \r\n");
	printf("  -b\r\n");
	printf("     output Big Endian (Motorola).\r\n");
	printf("\r\n");
	printf("  -h\r\n");
	printf("     Print this help.\r\n");
	printf("\r\n");
	printf("  -t\r\n");
	printf("     List known types.\r\n");
	printf("\r\n");
	printf("  -x\r\n");
	printf("     Use hex notation when possible.\r\n");
	printf("\r\n");
	printf("Victron Energy B.V. Jeroen Hofstee\r\n");
}

struct Error
{
	int ret;
	char const* str;

	Error(int ret, char const* str)
	{
		this->ret = ret;
		this->str = str;
	}
};

// add a new variant to the buffer
void convert(char const *str)
{
	un8 temp[30];

	VeVariant variant;

	variant.type.tp = veTypeFromName(str);
	if (variant.type.tp == VE_UNKNOWN)
		throw Error(-1, "# Invalid value format");

	if (bigEndian)
		veInVariantBe(&variant);
	else
		veInVariantLe(&variant);

	if (veStreamIn->error)
		throw Error(-1, "# Error underrun");

	veTypeName(variant.type, (char*) temp, sizeof(temp));
	printf("%s:", temp);
	if (isHex)
		veVariantFPrintX(stdout, &variant);
	else
		veVariantFPrint(stdout, &variant);
	printf("\r\n");
}

// main, parse arguments
int main(int argc, char* argv[])
{
	int n;
	int c;

#ifdef _MSC_VER
	// leave \r and \n alone!
	_setmode( _fileno( stdin ), _O_BINARY );
#endif

	while ((c = getopt(argc, argv, "bhtx")) != -1)
	{
		switch(c)
		{
			case 'b':
				bigEndian = true;
				break;

			case 'h':
				usage(argv[0]);
				return 0;

			case 'x':
				isHex = veTrue;
				break;

			case 't':
				char name[20];
				for (int datatype = 0; datatype <= 127; datatype++)
				{
					VeDatatype type;
					type.tp = (VeDataBasicType) datatype;
					type.len = 0;
					veTypeName(type, name, sizeof(name));
					printf("0x%02X %3d : %s\r\n", datatype, datatype, name);
				}
				return 0;
		}
	}

#if 1
	// binary
	n = 0;
	while(1)
	{
		c = fgetc(stdin);
		if (c == EOF)
			break;

		if (n >= (int) sizeof(buffer))
			throw Error(-1, "# Error data is too long");

		buffer[n] = (un8) c;
		n++;
	}
#else
	// hex reader
	n = 0;
	unsigned int number;
	int ret;
	while ((ret = scanf("0x%X ", &number)) > 0)
	{
		if (n >= sizeof(buffer))
		  throw Error(-1, "# Error data is too long");

		if (number > 255)
		{
			throw Error(-1, "# bytes only..");
		}

		buffer[n] = (un8) number;
		n++;
	}
#endif

	veInStreamInit(&stream, buffer, n);

	if (bigEndian)
		printf("# Big Endian\r\n");
	else
		printf("# Little Endian\r\n");

	// parse the arguments
	n = optind;
	try
	{
		while (n < argc)
		{
			convert(argv[n]);
			n++;
		}

		// force output of pending bits
		if (stream.bitsUsed != 0)
		{
			printf("# warning: bits left!\r\n");
			for (n = 0; n < stream.bitsUsed; n++)
				convert("bit1");
		}

		// force output of pending bytes
		if (veInBytesLeft() != 0)
		{
			printf("# warning: bytes left!\r\n");
			while (veInBytesLeft())
				convert("un8");
		}
	} catch (Error err)
	{
		fprintf(stderr, "%s", err.str);
		return err.ret;
	}

	return 0;
}
