// Simple tool for converting text to binary

#ifdef WIN32
#include <windows.h>
#include <io.h>
#endif

#include <velib/velib_config.h>
#include <velib/io/ve_vstream.h>
#include <velib/types/variant.h>
#include <velib/types/variant_console.h>
#include <velib/utils/ve_getopt.h>

#include <stdio.h>
#include <iostream>
#include <fcntl.h>
#include <string>
#include <string.h>

using namespace std;

un8 buffer[255];
VeStream stream;
bool bigEndian = false;

void usage(char* name)
{
	printf("%s [variant] [variant] ..\r\n", name);
	printf("\r\n");
	printf("Convert a textual representation of values to a binary.\r\n");
	printf("\r\n");
	printf("variant = type:value\r\n");
	printf("  type  - e.g. un8 see -t.\r\n");
	printf("  value - either numeric, hex e.g. 0xFF or string\r\n");
	printf("          by default the values are formatted as Little Endian (Intel)\r\n");
	printf("\r\n");
	printf("If no values are passed as argument, values are read from stdin.\r\n");
	printf("e.g. allowing %s -b < header.txt > header.bin\r\n", name);
	printf("Empty lines and lines starting with a # are ignored\r\n");
	printf("\r\n");
	printf("OPTIONS: \r\n");
	printf("  -b\r\n");
	printf("     output Big Endian (Motorola).\r\n");
	printf("\r\n");
	printf("  -h\r\n");
	printf("     Print this help.\r\n");
	printf("\r\n");
	printf("  -t\r\n");
	printf("     List known types.\r\n");
	printf("\r\n");
	printf("Victron Energy B.V. Jeroen Hofstee\r\n");
}

struct Error
{
	int ret;
	char const* str;

	Error(int ret, char const* str)
	{
		this->ret = ret;
		this->str = str;
	}
};

// add a new variant to the buffer
void convert(char* str)
{
	VeVariant variant;

	if (!veVariantFromDefStr(&variant, str))
		throw Error(-1, "Invalid value format");

	if (bigEndian)
		veOutVariantBe(&variant);
	else
		veOutVariantLe(&variant);

	if (veStreamOut->error)
		throw Error(-1, "Error formatting value");
}

// send the (whole byte) contents to the output
void output(void)
{
	for (int i=0; i < veStreamOut->index; i++)
	{
		fputc(buffer[i], stdout);
	}
	veStreamOut->index = 0;
}

// main, parse arguments
int main(int argc, char* argv[])
{
	int n;
	int c;

#ifdef _MSC_VER
	// leave \r and \n alone!
	_setmode( _fileno( stdout ), _O_BINARY );
#endif

	veOutStreamInit(&stream, buffer, sizeof(buffer));

	while ((c = getopt(argc, argv, "bht")) != -1)
	{
		switch(c)
		{
			case 'b':
				bigEndian = true;
				break;

			case 'h':
				usage(argv[0]);
				return 0;

			case 't':
				char name[20];
				VeDatatype tp;
				for (int datatype = 0; datatype <= 127; datatype++)
				{
					tp.tp = (VeDataBasicType) datatype;
					veTypeName(tp, name, sizeof(name));
					printf("0x%02X %3d : %s\r\n", datatype, datatype, name);
				}
				return 0;
		}
	}

	// parse the arguments
	n = optind;
	try
	{
		while (n < argc)
		{
			convert(argv[n]);
			output();
			n++;
		}
	} catch (Error err)
	{
		fprintf(stderr, "%s", err.str);
		return err.ret;
	}

	// force output of pending bits
	if (stream.bitsUsed != 0)
	{
		veOutBits8Le(0, 8 - stream.bitsUsed);
		output();
	}

	// if some values where provided as argument -> done
	if (optind != argc)
		return 0;

	// handle values from standard input
	while(!cin.eof())
	{
		string input;
		
		getline(cin, input);
		if (cin.eof())
			break;

		if (input.length() == 0 || input[0] == '#' || input[0] == '@')
			continue;

		// make a copy, since the string is actually changed
		char* buf = new char[input.length() + 1];
		strcpy(buf, input.c_str());
		try
		{
			convert(buf);
			output();
		}
		catch (Error err)
		{
			cerr << err.str << endl;
		}
		delete[] buf;
	}

	return 0;
}
