#ifdef _MSC_VER
#include <windows.h>
#include <io.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <fcntl.h>

#include <velib/base/base.h>
#include <velib/velib_inc_J1939.h>
#include <velib/platform/task.h>
#include <velib/platform/console.h>
#include <velib/canhw/canhw.h>
#include <velib/io/ve_stream.h>
#include <velib/io/ve_stream_n2k.h>
#include <velib/J1939/J1939.h>
#include <velib/J1939/J1939_device.h>
#include <velib/J1939/J1939_stack.h>
#include <velib/J1939/J1939_tx.h>
#include <velib/nmea2k/n2k.h>
#include <velib/vecan/vecan.h>
#include <velib/vecan/reg.h>
#include <velib/vecan/reg_prod_sf.h>
#include <velib/vecan/reg_rx.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_arg.h>
#include <velib/utils/ve_todo.h>

#include <velib/types/variant_console.h>
#include <velib/utils/ve_logger.h>

#define RET_OVERFLOW	10

typedef struct
{
	J1939Nad tg;
	VeRegId regId;
	VeRegId mask;
	veBool req;
} VRegInfo;

static veBool isBin;
static veBool keepRunning;
static veBool response;
static int mArgc;
static char **mArgv;
static veBool msgSend = veFalse;

un16 timeout;

VRegInfo vRegInfo = {0xFF, VE_REG_INVALID, 0xFFFF, veFalse};

struct option consoleOptions[] =
{
	{"bin", no_argument, 0, 'b'},
	{"get", required_argument, 0, 'g'},
	{"mask", required_argument, 0, 'm'},
	{"nad", required_argument, 0, 'n'},
	{"run", no_argument, 0, 'r'},
	{"set", required_argument, 0, 's'},
	{0,	0,	0,	0}
};

void veRegDataHandler(VeEvent ev);

VeEventHndlr const veRegRxHndlrDef[] =
{
	veRegDataHandler,
	NULL
};

void consoleUsage(char* program)
{
	printf("%s - Victron registers\r\n", program);
	printf("\r\n");
	printf("  -b,--bin\r\n");
	printf("   Output in binary format.\r\n");
	printf("\r\n");
	printf("  -g,--get\r\n");
	printf("   Get a register.\r\n");
	printf("\r\n");
	printf("  -n,--nad\r\n");
	printf("   The NAD of the target device.\r\n");
	printf("\r\n");
	printf("  -r,--run\r\n");
	printf("   Keep running.\r\n");
	printf("\r\n");
	printf("  -s,--set\r\n");
	printf("   Set a register. Data is set with e.g. un16:2800 un16:12\r\n");
	printf("\r\n");
	pltOptionsUsage();
	printf("Victron Energy B.V.\r\n");
}

veBool consoleOption(int flag)
{
	switch(flag)
	{
	case 'b':
		isBin = veTrue;
#ifdef _MSC_VER
		// leave \r and \n alone!
		_setmode( _fileno( stdout ), _O_BINARY );
#endif
		return veTrue;

	case 'g':
		vRegInfo.req = veTrue;
		errno = 0;
		vRegInfo.regId = veArgInt(optarg);
		if (errno == 0)
			return veTrue;

		printf("invalid register value");
		return veFalse;

		return veTrue;

	case 'm':
		errno = 0;
		vRegInfo.mask = veArgInt(optarg);
		if (errno == 0)
		{
			keepRunning = veTrue;
			return veTrue;
		}

		printf("invalid mask value");
		return veFalse;

	case 'n':
		errno = 0;
		vRegInfo.tg = veArgInt(optarg);
		if (errno == 0)
			return veTrue;

		printf("invalid nad value");
		return veFalse;

	case 'r':
		keepRunning = veTrue;
		return veTrue;

	case 's':
		vRegInfo.req = veFalse;
		errno = 0;
		vRegInfo.regId = veArgInt(optarg);
		if (errno == 0)
			return veTrue;

		printf("invalid register value");
		return veFalse;

	}

	printf("invalid flag");
	return veFalse;
}

veBool consoleArgs(int argc, char *argv[])
{
	if (vRegInfo.tg == J1939_NAD_BROADCAST && !isBin)
		printf("# reminder, there are no error indications when broadcasting!\r\n");
	
	mArgc = argc;
	mArgv = argv;

	return veTrue;
}

void taskInit(void)
{
	J1939Sf sf;
	j1939Device.nad = 0xFE;
	j1939Device.flags |= J1939_ACL_VALID_NAD;
	timeout = 0;

	while(veCanRead((VeRawCanMsg*) &sf));
}

void taskUpdate(void)
{
	J1939Sf sf;
	J1939Msg* msg;

	while(veCanRead((VeRawCanMsg*) &sf))
	{
		if (!j1939MsgForDevice(&(sf.msgId)))
			continue;

		veInMsg((J1939Msg*) &sf);
		if (sf.msgId.p.pg.p.dp == 0 && J1939PgnFromIdNoTg(sf.msgId) == PGN0_VECAN_VREG)
		{
			if (veInUn16Le() != N2K_MAN_TAG_VICTRON)
				continue;
			veRegRxData();
		}		
	}

	while ((msg = j1939TxPop()) != NULL)
	{
		veAssert(msg->type.p.kind == J1939_SF);
		veCanSend((VeRawCanMsg*) msg);
		j1939SfTxFree((J1939Sf*) msg);
	}
}

// Request for normal mode
void veRegDataHandler(VeEvent ev)
{
	if (vRegInfo.tg != J1939_NAD_BROADCAST && n2kMsgIn->msgId.p.src != vRegInfo.tg)
		return;

	if (!msgSend)
		return;

	if ((veRegRxId & vRegInfo.mask) == (vRegInfo.regId & vRegInfo.mask))
	{
		switch (ev)
		{
		case VE_EV_VREG_RX:
			response = veTrue;
			if (isBin)
			{
				while (veInBytesLeft())
					printf("%c", veInUn8());
			} else {
				if (!isBin)
					printf("# NAD 0x%X, reg=0x%X\r\n", n2kMsgIn->msgId.p.src, veRegRxId);
				while (veInBytesLeft())
					printf("0x%02X ", veInUn8());
				printf("\r\n");
			}
			if (!keepRunning)
				pltExit(0);
			timeout = 0;
			break;

		case VE_EV_VREG_RX_ACK:
			response = veTrue;

			if (!isBin)
			{
				printf(
					"# ERROR NAD 0x%X, reg=0x%X code=0x%X\r\n",
					n2kMsgIn->msgId.p.src, 
					veRegRxId,
					veRegAckCode
				);

				switch (veRegAckCode & 0xFF00)
				{
				case VACK_ERR_REQ:
					printf("# register cannot be requested %X\r\n", veRegAckCode);
					break;

				case VACK_ERR_CMD:
					printf("# register cannot be changed %X\r\n", veRegAckCode);
					break;
				}
			}
			if (!keepRunning)
				pltExit(1);
			timeout = 0;
			break;
		
		default:
			;
		}
	}

	if (ev == VE_EV_VREG_RX_ACK) 
	{
		response = veTrue;
		switch(veRegAckCode)
		{
		case VACK_ERR_OVERFLOW:
			fprintf(stderr, "#Received error: too many invalid request! quiting..\r\n");
			if (!keepRunning)
				pltExit(RET_OVERFLOW);
			break;
		}
	}
}

void taskTick(void)
{
	if (timeout++ > 5 * 20)
	{
		if (!response)
			fprintf(stderr, "# Error no response received!\r\n");
		if (!keepRunning)
			pltExit(0);
	}

	if (msgSend)
		return;
	
	if (vRegInfo.req)
	{
		if (veRegRequestMask(vRegInfo.tg, vRegInfo.regId, vRegInfo.mask))
			msgSend = veTrue;
	} else {
		int n;
		veRegPrepareSf(vRegInfo.tg, vRegInfo.regId);
		if (n2kMsgOut == NULL)
			return;

		for (n = 0; n < mArgc; n++) {
			VeVariant variant;
			if (!veVariantFromDefStr(&variant, mArgv[n])) {
				fprintf(stderr, "#invalid format %s\r\n", mArgv[n]);
				pltExit(-1);
			}
			veOutVariantLe(&variant);
		}

		if (veStreamOut->error) {
			fprintf(stderr, "#variables didn't fit\r\n");
			pltExit(-1);
		}

		veOutSendSf();
		msgSend = veTrue;
	}
}
